/* THIS FILE IS AUTOGENERATED - DO NOT EDIT */

#ifndef mozilla_dom_MobileMessageManagerBinding_h__
#define mozilla_dom_MobileMessageManagerBinding_h__

#include "js/RootingAPI.h"
#include "jspubtd.h"
#include "mozilla/ErrorResult.h"
#include "mozilla/dom/BindingDeclarations.h"
#include "mozilla/dom/Nullable.h"

namespace mozilla {
namespace dom {

struct MmsAttachment;
struct MmsParametersAtoms;
class NativePropertyHooks;

} // namespace dom
} // namespace mozilla

namespace mozilla {
namespace dom {

struct MmsParameters : public MainThreadDictionaryBase
{
  Optional<Sequence<MmsAttachment > > mAttachments;
  Optional<Sequence<nsString > > mReceivers;
  nsString mSmil;
  nsString mSubject;

  MmsParameters();

protected:
  explicit inline MmsParameters(int )
  {
    // Do nothing here; this is used by our "Fast" subclass
  }

private:
  MmsParameters(const MmsParameters&) MOZ_DELETE;
  void operator=(const MmsParameters) MOZ_DELETE;

  static bool
  InitIds(JSContext* cx, MmsParametersAtoms* atomsCache);

public:
  bool
  Init(JSContext* cx, JS::Handle<JS::Value> val, const char* sourceDescription = "Value");

  bool
  Init(const nsAString& aJSON);

  bool
  ToObject(JSContext* cx, JS::Handle<JSObject*> parentObject, JS::MutableHandle<JS::Value> rval) const;

  void
  TraceDictionary(JSTracer* trc);
};

namespace dictionary_detail {
struct FastMmsParameters : public MmsParameters
{
  inline FastMmsParameters()
    : MmsParameters(42)
  {
    // Doesn't matter what int we pass to the parent constructor
  }
};
} // namespace dictionary_detail


} // namespace dom
} // namespace mozilla

#endif // mozilla_dom_MobileMessageManagerBinding_h__
